# Copyright (c) Microsoft Corporation
# Licensed under the MIT License.

import shutil
from pathlib import Path
from tempfile import TemporaryDirectory

import PIL
import pytest
from common_vision_utils import (DummyFlowersPipelineSerializer,
                                 ImageClassificationPipelineSerializer,
                                 create_dummy_model,
                                 create_image_classification_pipeline,
                                 load_flowers_dataset, load_imagenet_dataset,
                                 load_imagenet_labels)
from rai_vision_insights_validator import run_and_validate_serialization

from responsibleai_vision import ModelTask, RAIVisionInsights
from responsibleai_vision.common.constants import ImageColumns


class TestRAIVisionInsightsSaveAndLoadScenarios(object):

    def test_rai_insights_empty_save_load_save(self):
        data = load_imagenet_dataset()
        pred = create_image_classification_pipeline()
        task_type = ModelTask.IMAGE_CLASSIFICATION
        class_names = load_imagenet_labels()
        test = data[:3]
        label = ImageColumns.LABEL
        serializer = ImageClassificationPipelineSerializer()

        run_and_validate_serialization(
            pred, test, task_type, class_names, label, serializer)

    @pytest.mark.skip("Insufficient memory on test machines to load images")
    def test_rai_insights_large_images_save_load_save(self):
        PIL.Image.MAX_IMAGE_PIXELS = None
        data = load_flowers_dataset(upscale=True)
        model = create_dummy_model(data)
        test = data
        class_names = data[ImageColumns.LABEL.value].unique()
        task_type = ModelTask.IMAGE_CLASSIFICATION
        label = ImageColumns.LABEL
        serializer = DummyFlowersPipelineSerializer()

        image_width = 2
        run_and_validate_serialization(
            model, test, task_type, class_names, label, serializer,
            image_width)

    def test_loading_rai_insights_without_model_file(self):
        data = load_imagenet_dataset()
        pred = create_image_classification_pipeline()
        task_type = ModelTask.IMAGE_CLASSIFICATION
        class_names = load_imagenet_labels()
        test = data[:3]
        label = ImageColumns.LABEL
        serializer = ImageClassificationPipelineSerializer()

        rai_insights = RAIVisionInsights(
            pred, test, label,
            task_type=task_type,
            classes=class_names,
            serializer=serializer)

        with TemporaryDirectory() as tmpdir:
            assert rai_insights.model is not None
            save_path = Path(tmpdir) / "rai_insights"
            rai_insights.save(save_path)

            # Remove the model.pkl file to cause an exception to occur
            # while loading the model.
            model_name = 'image-classification-model'
            model_pkl_path = Path(tmpdir) / "rai_insights" / model_name
            shutil.rmtree(model_pkl_path)
            match_msg = 'No file or directory found'
            with pytest.raises(OSError, match=match_msg):
                without_model_rai_insights = RAIVisionInsights.load(save_path)
                assert without_model_rai_insights.model is None



# Copyright (c) Microsoft Corporation
# Licensed under the MIT License.

"""Contains image handling utilities."""

import numpy as np

from responsibleai_vision.common.constants import ImageColumns
from responsibleai_vision.utils.image_reader import get_image_from_path

IMAGE = ImageColumns.IMAGE.value
IMAGE_URL = ImageColumns.IMAGE_URL.value
IMAGE_DETAILS = 'image_details'
LABEL = 'label'
WIDTH = 'width'
HEIGHT = 'height'
TOP_X = 'topX'
TOP_Y = 'topY'
BOTTOM_X = 'bottomX'
BOTTOM_Y = 'bottomY'
IS_CROWD = 'isCrowd'


def convert_images(dataset, image_mode):
    """Converts the images to the format required by the model.

    If the images are base64 encoded, they are decoded and converted to
    numpy arrays. If the images are already numpy arrays, they are
    returned as is.

    :param dataset: The dataset to convert.
    :type dataset: numpy.ndarray
    :param image_mode: The mode to open the image in.
        See pillow documentation for all modes:
        https://pillow.readthedocs.io/en/stable/handbook/concepts.html
    :type image_mode: str
    :return: The converted dataset.
    :rtype: numpy.ndarray
    """
    if len(dataset) > 0 and isinstance(dataset[0], str):
        try:
            dataset = np.array([get_image_from_path(
                x, image_mode) for x in dataset])
        except ValueError:
            # if images of different sizes, try to convert one by one
            jagged = np.empty(len(dataset), dtype=object)
            for i, x in enumerate(dataset):
                jagged[i] = get_image_from_path(x, image_mode)
            dataset = jagged
    return dataset


def get_images(dataset, image_mode, transformations=None):
    """Get the images from the dataset.

    If transformations are provided as a callable, the images
    are transformed. If transformations are provided as a string,
    the images are retrieved from that column name in the test dataset.

    :param dataset: The dataset to get the images from.
    :type dataset: numpy.ndarray
    :param image_mode: The mode to open the image in.
        See pillow documentation for all modes:
        https://pillow.readthedocs.io/en/stable/handbook/concepts.html
    :type image_mode: str
    :param transformations: The transformations to apply to the images.
    :type transformations: torchvision.transforms
    :return: The images.
    :rtype: numpy.ndarray
    """
    column_names = dataset.columns
    is_transformations_str = isinstance(transformations, str)
    if is_transformations_str:
        images = dataset[transformations]
    else:
        if IMAGE in column_names:
            images = dataset[IMAGE]
        elif IMAGE_URL in column_names:
            images = dataset[IMAGE_URL]
        else:
            raise ValueError('No image column found in test data')

    images = np.array(images.tolist())
    converted_images = convert_images(images, image_mode)

    if not is_transformations_str and transformations is not None:
        converted_images = transformations(converted_images)

    return converted_images


def classes_to_dict(classes):
    """Converts the classes to a dictionary.

    :param classes: The classes.
    :type classes: list
    :return: The classes as a dictionary.
    :rtype: dict
    """
    return {classes[i]: i + 1 for i in range(len(classes))}


def transform_object_detection_labels(test, target_column, classes):
    """Transforms the object detection labels to one common format.

    :param test: The test dataset.
    :type test: pandas.DataFrame
    :param target_column: The column containing the labels.
    :type target_column: str
    """
    label_dict = classes_to_dict(classes)
    for i in range(len(test)):
        object_labels = test[target_column][i]
        image_details = None
        if IMAGE_DETAILS in test:
            image_details = test[IMAGE_DETAILS][i]
        if len(object_labels) > 0 and isinstance(object_labels[0], dict):
            if image_details:
                width = image_details[WIDTH]
                height = image_details[HEIGHT]
                image_labels = []
                for label in object_labels:
                    class_name = label[LABEL]
                    class_id = label_dict[class_name]

                    xmin = label[TOP_X] * width
                    ymin = label[TOP_Y] * height

                    xmax = label[BOTTOM_X] * width
                    ymax = label[BOTTOM_Y] * height

                    image_labels.append([class_id, int(xmin), int(ymin),
                                         int(xmax), int(ymax),
                                         int(label[IS_CROWD])])
                test[target_column][i] = image_labels
            else:
                invalid_msg = 'Invalid label format for conversion: '
                err = invalid_msg + 'Image details and label must be present'
                raise ValueError(err)
    return test



# Copyright (c) Microsoft Corporation
# Licensed under the MIT License.

from math import isclose

import numpy as np
from common_vision_utils import load_fridge_object_detection_dataset

from responsibleai_vision.common.constants import ImageColumns
from responsibleai_vision.utils.image_utils import (
    BOTTOM_X, BOTTOM_Y, HEIGHT, IS_CROWD, TOP_X, TOP_Y, WIDTH, classes_to_dict,
    transform_object_detection_labels)

LABEL = ImageColumns.LABEL.value
IMAGE_DETAILS = ImageColumns.IMAGE_DETAILS.value
TOL = 1


class TestImageUtils(object):
    def test_transform_object_detection_labels(self):
        data = load_fridge_object_detection_dataset(automl_format=True)
        class_names = np.array(['can', 'carton',
                                'milk_bottle', 'water_bottle'])
        data_transformed = transform_object_detection_labels(
            data.copy(), LABEL, class_names)
        label_dict = classes_to_dict(class_names)
        for i in range(len(data)):
            original_label = data[LABEL][i]
            label = data_transformed[LABEL][i]
            image_details = data[IMAGE_DETAILS][i]
            width = image_details[WIDTH]
            height = image_details[HEIGHT]
            assert isinstance(label, list)
            for j in range(len(label)):
                label_j = label[j]
                assert len(label_j) == 6
                o_label_j = original_label[j]
                assert label_j[0] == label_dict[o_label_j[LABEL]]
                # use is close for off by one errors due to rounding
                assert isclose(label_j[1], o_label_j[TOP_X] * width,
                               abs_tol=TOL)
                assert isclose(label_j[2], o_label_j[TOP_Y] * height,
                               abs_tol=TOL)
                assert isclose(label_j[3], o_label_j[BOTTOM_X] * width,
                               abs_tol=TOL)
                assert isclose(label_j[4], o_label_j[BOTTOM_Y] * height,
                               abs_tol=TOL)
                assert label_j[5] == o_label_j[IS_CROWD]
